/*
   author:  Rony G. Flatscher
   date:    2005-04-18/19/20
   todo:     - loadClass -> use bsf.import of the class (so newStrict is there)
             - try to allow setCell to be used with named cells
             - add a getCell(Range) ?
   changed:
            2006-06-04, ---rgf, fixed a bug in querying an interface object, if unqualified class name is
                                used more than once
            2006-02-03, ---rgf, commented fall-back-strategy to query interface object via RgfReflectUNO-class
                                in UNO_PROXY.UNKNOWN
            2006-01-06, ---rgf, removed public routines pp() and iif(), because they are now availble
                                in the required BSF.CLS module; added the public routines
                                uno.getScriptContext() and uno.getScriptContextVersion()

            2005-05-03, ---rgf, made sure that exceptions are propagated
            2005-06-05, ---rgf, uppercasing "OOo.cls" to "OOO.CLS", removing
            2005-07-03, ---rgf, changed logic to queryInterface(), now dealing with duplicate class names
                                gracefully; fixed a bug (do not wrap up BSF arrays)
            2005-07-04, ---rgf, optimized handling interface classes,
                                added OOo.getInterfaces(object[, sortByUnqualifiedClassName]),
                                added OOo.getServices(object)
            2005-07-05, ---rgf, removed unnecessary preloading of application classes
            2005-07-08, ---rgf, ooo.wrap(): now tries a bsf.wrap(), if ncecessary
            2005-07-30, ---rgf, adapted OOO.CLS to new array-behaviour in BSF.CLS
            2005-08-18, ---rgf, added public routines "ConvertToURL()", "ConvertFromURL()", which use
                                "encodeURL()", "decodeURL()"
            2005-08-31, ---rgf, renamed module to "UNO.CLS" and renamed stuff from "OOO" to "UNO" to
                                reflect the generic support for UNO; office is just one combination of
                                UNO components and UNO application; added license
            2005-12-15, ---rgf, added newly developed Java class "org.oorexx.uno.RgfReflectUNO",
                                added entry "version" in ".uno"
            2005-12-16, ---rgf, added wrapper routines for RgfReflectUNO methods which wrap returned
                                UNO objects into ooRexx UNO_PROXY objects;
            2005-12-21, ---rgf, adapted and added wrapper routines for RgfReflectionUNO
            2005-12-26, ---rgf, adapted UNO-routine names to new RgfReflectionUNO names, added
                                UNO.*-methods to UNO_PROXY
            2005-12-27, ---rgf, updated license part, added utility class "UNO_ENUM_OR_CONSTANTS"
            2005-12-28, ---rgf, generalized the "directoryfiying" of UNOIDL definitions for "UNO_ENUM",
                                "UNO_CONSTANTS" and "UNO_PROPERTIES" (removed "UNO_ENUM_OR_CONSTANTS")
            2005-12-29, ---rgf, "UNO_ENUM" now returns the Java object representing the enum-value,
                                "UNO_CONSTANTS" will return the constant's value (a number)
            2006-01-03, ---rgf, corrected inline documentation


   version: 1.0.2
   name:    UNO.CLS
   purpose: allow interfacing with UNO components (via Java), tested with OpenOffice.org, v1.1.x and up
   needs:   BSF4Rexx, BSF.CLS, UNO_XINTERFACES.REX

   license:

    ------------------------ Apache Version 2.0 license -------------------------
       Copyright (C) 2006 Rony G. Flatscher

       Licensed under the Apache License, Version 2.0 (the "License");
       you may not use this file except in compliance with the License.
       You may obtain a copy of the License at

           http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing, software
       distributed under the License is distributed on an "AS IS" BASIS,
       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
       See the License for the specific language governing permissions and
       limitations under the License.
    -----------------------------------------------------------------------------
*/


   -- if no environment entry for UNO as of yet, then create the entry with default classes
if .local~hasentry("UNO")=.false then
do
   .local~UNO=.directory~new     -- create a new directory object and store it in .local
   .uno~version="102.20060604"   -- set version number
   .uno~bExtendSearch=.true      -- as of 2005-12-20: if interface not found via XTypeProvider,
                                 -- use full reflection (slow, but there seem to be bugs in XTypeProvider)

   -- preload some important classes, just send their class name to the directory object ".UNO"
   l = .list~of( "com.sun.star.beans.PropertyValue"   , -
                 "org.oorexx.uno.RgfReflectUNO"       , -
                 "com.sun.star.comp.helper.Bootstrap" , -
                 "com.sun.star.uno.Any"               , -
                 "com.sun.star.uno.AnyConverter"      , -
                 "com.sun.star.uno.UnoRuntime"       )

   do c over l -- get Java class object and save it in .UNO by className
      call UNO.loadClass c
   end

-- trace ?all
-- trace ?i

      -- create empty array of property-value (needed quite often)
   .UNO~noProps=.bsf~bsf.createArray(.UNO~propertyValue, 0)

      -- list of Interface classes to be loaded at runtime (needs completion)
   l = UNO.getInterfaceList()    -- retuns > 950 UNO-interface classes in a list

   -- save interface class names with which we want to deal
   d=.directory~new  -- directory for all InterfaceTypes
   r=.relation~new   -- relation to store duplicates (if last word is used multiple times, keep full class name)
   do i over l
     idx=substr(i, lastpos(".", i)+1)
     d~setentry(i, i)    -- by fully qualified name

     if d~hasentry(idx) then  -- interface class name used already, use relation to store full name
     do
        r[idx~translate]=i    -- save full name with class name
     end
     else   -- new interface class name
     do
       d~setentry(idx, i)     -- by last word
     end
   end
   .UNO~XInterfaces = d       -- save directory
   .UNO~XInterfaces.dupes = r -- save relation
end


::requires BSF.CLS                -- get BSF4Rexx support
::requires UNO_XINTERFACES.REX    -- get the list of all of UNO XInterface classes


      -- load UNO-class and save it in environment; "idx" determines the name which is used to
      -- store the class object in the directory ".UNO"
::routine UNO.loadClass public
     parse arg className, idx

     if idx="" then idx=substr(className, lastpos(".", className)+1) -- extract last word to serve as index
     if .UNO~hasentry(idx)=.true then    -- e.g. last word is used as a class in a another package as well
        .error~say("---> UNO.loadClass():" pp(idx) "already used, in hand:" pp(className) "in .UNO:" pp(.UNO~entry(idx)))

 -- say "UNO.CLS: className="className "idx="idx
 -- say "UNO.CLS:    imported .bsf~bsf.import(className, .nil)=" .bsf~bsf.import(className, .nil)

     tmpClass=.UNO_proxy~new(.bsf~bsf.import(className, .nil))  -- make it look like an ooRexx class
 -- say "UNO.CLS:    imported .Uno_proxy(...)=" tmpClass

     .UNO~setentry(idx, tmpClass)   -- load the given Java class and save it in .UNO
     return tmpClass


   /*
      connects to the UNO runtime environment and returns the XContext object; if passing a URL,
      then the remote object is returned
   */
::routine UNO.connect public
  parse arg unoURL

  xContext=.UNO~Bootstrap~bootstrap          -- get an initialized local UNO runtime environment

  if unoURL="" then  -- no unoURL given, use default (local installation)
  do
     return xContext       -- return the context object (which can be used to get the service manager)
  end

  uur= xContext~getServiceManager~createInstanceWithContext("com.sun.star.bridge.UnoUrlResolver", xContext)
  remoteObject = uur~XUnoUrlResolver ~resolve(unoURL)    -- get the connection
  return remoteObject

/*
 -- unoURL="uno:socket,host=localhost,port=8100;urp;StarOffice.ServiceManager"
    remoteMSF = remoteServiceManager~XMultiServiceFactory

--- or:
--  unoURL="uno:socket,host=localhost,port=8200;urp;StarOffice.NamingService"
  remoteNamingService = remoteObject~XNamingService

  remoteServiceManager=remoteNamingService~getRegisteredObject("StarOffice.ServiceManager")
  remoteMSF           = remoteServiceManager~XMultiServiceFactory
  if .UNO~hasentry("remoteMSF")=.false then .UNO~remoteMSF=remoteMSF -- save remote MultiServiceFactory
  return remoteMSF  -- return the remote XMultiServiceFactory
*/


      -- get a XContext, use it to create and to return the XDesktop interface object
::routine UNO.createDesktop public
  USE ARG xContext

  if arg()=0 | .nil=xContext then   -- no context given, create default (local) one
     xContext=UNO.connect()

  service="com.sun.star.frame.Desktop" -- OOo Desktop UNO component service
  return xContext~getServiceManager~createInstanceWithContext(service, xContext)


   -- a routine for inserting values or formulas into cells
::routine UNO.setCell public

   use arg xSheet, x, y, content
	xSheet~getCellByPosition(x, y) ~setFormula(content)



/* *************************************************************
   Class:       UNO_PROXY
   Purpose:     allows to wrap up Java objects as Object Rexx objects;
                if instances are destroyed, then no derigstration from
                the BSF registry takes place
   restriction: needed to make getting interface objects easier
   ************************************************************* */
::CLASS UNO_PROXY -- subclass BSF PUBLIC

::Method bsfObject ATTRIBUTE

::METHOD init
   expose arg bsfObject -- direct access to BSF object
   use arg bsfObject    -- assign object to attribute

      -- set beanName
   self~objectName=bsfObject~objectName   -- use same objectname as BSF proxy (= idx into BSF-registry)

      -- if bsfObject is an imported Java class object, then use its beanName directly
   if bsfObject~string~left(4)~translate="THE " then
   do
      signal on syntax name continue   -- if an error, not a JCO, hence use its "plain" objectname
      self~objectname=.bsf~bsf.getJavaClassObjectProxy(bsfObject~bsf.FullJavaClassName)~string
   end
   return

continue:   -- catch exception
/*
   -- default objectname
   self~objectName=bsfObject~objectName   -- use same objectname as BSF proxy (= idx into BSF-registry)
say "UNO.CLS ---> ---> self~objectName="self~objectname
*/


::METHOD unknown
   expose arg bsfObject -- direct access to BSF object
   parse upper arg xName 1 idx   -- make sure message name is in uppercase (if quoted)

   if pos(".", idx)>0 then idx=substr(idx, lastpos(".", idx)+1)   -- getindex

   SIGNAL ON SYNTAX NAME ANY  -- intercept all exceptions
   if idx~left(1)="X" then    -- if an interface object is asked for, query and return it
   do
      xClass=getInterfaceClass(xName, idx)

         -- query and return the interface object
      interface=.UNO~unoRuntime~queryInterface(xClass, bsfObject)
      if interface=.nil then  -- not found, maybe a duplicate simple class name ?
      do
         do fullName over .UNO~XInterfaces.dupes~allAt(idx)
            idx=substr(fullName, lastpos(".", fullName)+1)   -- getindex
            xClass=getInterfaceClass(fullName, idx)      -- rgf, 2006-06-04

            if .nil<>xClass then
               interface=.UNO~unoRuntime~queryInterface(xClass, bsfObject)

            if .nil<>interface then
               return .UNO_proxy~new(interface)
         end

         -- try the new RgfReflectUNO, maybe a new interface, not in UNO_XINTERFACES.REX
         -- as of 2006-02 commented: if this fall-back is used while running via OOo's Java scripting framework
         -- it would hang that framework hard!
         -- return .UNO_proxy~new(.uno~RgfReflectUno~queryInterfaceObjectByName(bsfObject, xName, .uno~bExtendSearch ))

         return .nil          -- no interface available
      end

      return .UNO_proxy~new(interface)
      -- return .UNO_proxy~new(.UNO~unoRuntime~queryInterface(xClass, bsfObject))
   end
   else  -- pass to BSF to handle it
   do
      if bsfObject~hasmethod(xName) then -- o.k. it's a BSF method (could be ooRexx OBJECT method as well!)
         FORWARD TO (bsfObject) MESSAGE (xName) ARGUMENTS (arg(2))
      else
         FORWARD TO (bsfObject) CONTINUE -- let BSF handle this

      tmpObj=result

      -- if a BSF-object, but not a BSF array: wrap it up
      if tmpObj~hasmethod("bsf.setFieldValue") then
      do
           -- if a BSF_ARRAY, then leave unchanged
         tmpClass=tmpObj~class~id   -- get class name
         if abbrev(tmpClass, "BSF_ARRAY") then return tmpObj   -- do not wrap up a BSF array object

           -- plain BSF object, hence wrap it up
         return .UNO_proxy~new(tmpObj)
      end

      return tmpObj
   end

ANY:
-- say "---> ---> ---> UNOps! UNO.CLS, SIGL="SIGL "condition()="condition()
  RAISE PROPAGATE    -- raise exception in caller

-- local function
-- try to get Interface class object; if not available, load it ("lazy loading")
getInterfaceClass: procedure
   parse arg xName, idx

   if xName<>idx then xClass=.UNO~entry(xName)  -- try to get interface class object, if already created
                 else xClass=.UNO~entry(idx)    -- try to get interface class object, if already created

   if xClass=.nil then        -- no entry as of yet
   do
      if xName<>idx then      -- long name given, then use that
      do
         xClass=UNO.loadClass(.UNO~XInterfaces~entry(xName), xName)
         if xClass<>.nil then -- o.k. also save with class name
         do
             .UNO~setEntry(idx, xClass)
-- say "long class name" pp(xName)"/"pp(idx)
         end
      end
      else     -- only class name is given
      do
         xClass=UNO.loadClass(.UNO~XInterfaces~entry(idx), idx)
         if xClass<>.nil then -- o.k., also save with full class name
         do
             .UNO~setEntry(.UNO~XInterfaces~entry(idx), xClass)
-- say "class name" pp(idx)"/"pp(.UNO~XInterfaces~entry(idx))
         end
      end
   end

   return xclass



   -- interface to UnoRuntime.areSame(o1, o2)
::method uno.isSame
   use arg o
   return .uno~UnoRuntime~areSame(self, o) /* true if both reference the same UNO object   */


-- supply the UNO.-routines in form of methods as well, makes it easier to use them

   /* interface to RgfReflectUNO.findInterfaceWithMember(): can be a string or an object
      if "bReturnString=.false", then carry out a queryInterface() and return that object (or
      .nil); if a string should be returned ("bReturnString=.true"), then "howMany" determines
      after how many matches the search should be ended (default "-1", ie. try to find all
      matches)
   */
::method uno.findInterfaceWithMember
   use arg name, bReturnString, howMany
   return uno.findInterfaceWithMember(self, name, bReturnString, howMany)


   -- interface to RgfReflectUNO.getDefinition(): returns always a blank delimited string
::method uno.getDefinition
   return uno.getDefinition(self)


   -- interface to RgfReflectUNO.getInterfacesViaReflection(): returns always a blank delimited string
::method uno.getInterfaceNamesViaReflection
   return uno.getInterfaceNamesViaReflection(self)


   -- interface to RgfReflectUNO.getProperties(): returns always a blank delimited string
::method uno.getProperties
   return uno.getProperties(self)


   -- interface to RgfReflectUNO.getServicesViaReflection(): returns always a blank delimited string
::method uno.getServiceNamesViaReflection
   return uno.getServiceNamesViaReflection(self)


   -- interface to RgfReflectUNO.getType(): returns always a string
::method uno.getTypeName
   return uno.getTypeName(self)


   -- interface to RgfReflectUNO.getXTypeProviderTypes(): returns always a blank delimited string
::method uno.getXTypeProviderTypeNames
   return uno.getXTypeProviderTypeNames(self)


   -- interface to RgfReflectUNO.queryInterfaceName(): returns always a string
::method uno.queryInterfaceName
   use arg name
   return uno.queryInterfaceName(self, name)


   -- interface to RgfReflectUNO.queryInterfaceObjectByName(): returns the interface object or .nil
::method uno.queryInterfaceObjectByName
   use arg name
   return uno.queryInterfaceObjectByName(self, name)


   -- interface to RgfReflectUNO.queryServiceName(): returns always a string
::method uno.queryServiceName
   use arg name
   return uno.queryServiceName(self, name)





/* *************************************************************
   Class:      UNO_DIRLIKE
   Purpose:    allows to make all UNO definitions for ENUM, CONSTANTS and PROPERTIES
               available via an ooRexx directory object, easying lookups considerably
   ************************************************************* */
::CLASS UNO_DIRLIKE SUBCLASS Directory
::method unoidl_name       attribute   -- fully qualified name of the enum/constant groupe
::method unoidl_definition attribute   -- UNOIDL definition in form of a blank delimited string
::method unoidl_typeName   attribute   -- UNOIDL typeName, will be .nil, if setup could not be carried out
::method unoidl_className  attribute   -- the fully qualified Java class name representing
::method init
       -- establish access to object attributes
    expose unoidl_name unoidl_definition unoidl_typeName unoidl_className
    parse arg unoidl_name              -- parse argument as string

    unoidl_typeName=uno.getTypeName(unoidl_name)      -- get UNO type name

    if self~setup=.false then  -- something went wrong, indicate it
    do
       -- drop unoidl_name unoidl_typeName               -- undefine
       unoidl_typeName=.nil            -- indicates that no content could be created
       unoidl_className=.nil
       unoidl_definition=pp(unoidl_name) "could not be successfully processed for this class" pp(self~class~string)
       return
    end



/* USE BSF-routine "bsf.wrapStaticFields(strJavaClassName)" instead (available as of 2006-02-03) */

/* *************************************************************
   Class:      UNO_ENUM
   Purpose:    make all UNO_ENUMS available via an ooRexx directory object
   Remark:     - there will be an additional entry "defaultEnumValue" created
               - Java-UNO expects enumeration values to be Java objects, which
                 are stored with the constant fields of the Java class representing
                 the UNOIDL-ENUM
               - sending the enum name or its numeric value will return the
                 appropriate Java-enum-object
   ************************************************************* */
::CLASS nixi_UNO_ENUM SUBCLASS UNO_DIRLIKE -- PUBLIC
::method setup       -- check and setup directory object

    if self~unoidl_typeName<>"UNO_ENUM" then          -- of correct UNO type ?
       return .false

    self~unoidl_definition=uno.getDefinition(self~unoidl_name)  -- get the UNOIDL-definition

         -- extract and save the (Java) className
    parse value self~unoidl_definition with . "|" className "|" .
    self~unoidl_className=className

    tmpArr=self~unoidl_definition~makearray(" ")

    parse value tmpArr[1] with . "|" . "|" val rest   -- ENUM: default value, CONSTANTS: typeClass
    self~setentry("default_Enum_Value", val)          -- make default enumeration value available

    do i=2 to tmpArr~items                            -- iterate over ENUM values
       parse value tmpArr[i] with name "|" val
       self~setentry(name, val)
       self~setentry(val, name)  -- now have the number value point to the case sensitive name
    end
    return .true                                      -- everything worked out o.k.

::method at          -- intercept AT-message
   parse arg val
--    say "   " self~string"/"self~"=="~c2x": AT()-message,   arg(1)="pp(val)

   if val="DEFAULT_ENUM_VALUE" then       -- get the enum-field's name representing the default-value
      val=self~at:super(self~at:super(val))

   if datatype(val, "Num")=.true then     -- a number in hand, get its corresponding enum-name
      val=self~at:super(val)

   -- retrieve Java object representing enum-value and return it
   return bsf.getConstant(self~unoidl_className, val)

   -- forward class (super)

::method "[]"        -- intercept []-message
--   say "   " self~string"/"self~"=="~c2x": []()-message,   arg(1)="pp(val)
   forward message ("AT")        -- re-route to "AT"-method; just to be sure that "[]" is treated
                                 -- in any case like "AT"

::method unknown
--   say "   " self~string"/"self~"=="~c2x": UNKNOWN()-message,   arg(1)="pp(arg(1)) "arg(2)="pp(arg(2))
   forward class (super) continue
--   say "   " "     result="pp(result)
   return self~at(result)        -- now let local AT-method do its thingy ...




/* USE BSF-routine "bsf.wrapStaticFields(strJavaClassName)" instead (available as of 2006-02-03) */

/* *************************************************************
   Class:      UNO_CONSTANTS
   Purpose:    make all UNO_CONSTANTS name/value pairs available via an ooRexx directory object
   Remark:     - sending the constant name will return its (numeric) value
   ************************************************************* */
::CLASS nixi_UNO_CONSTANTS SUBCLASS UNO_DIRLIKE -- PUBLIC
::method setup       -- check and setup directory object

    if self~unoidl_typeName<>"UNO_CONSTANTS" then     -- of correct UNO type ?
       return .false

    self~unoidl_definition=uno.getDefinition(self~unoidl_name)  -- get the UNOIDL-definition

         -- extract and save the (Java) className
    parse value self~unoidl_definition with . "|" className "|" .
    self~unoidl_className=className

    tmpArr=self~unoidl_definition~makearray(" ")

    do i=2 to tmpArr~items                            -- iterate over CONSTANT values
       parse value tmpArr[i] with name "|" val
       self~setentry(name, val)
       self~setentry(val, name)  -- now have the number value point to the case sensitive name
    end
    return .true                                      -- everything worked out o.k.

/*   /* the following methods would make sure that a java.lang.Long object will be returned */
::method at          -- intercept AT-message
   parse arg val
--    say "   " self~string"/"self~"=="~c2x": AT()-message,   arg(1)="pp(val)

   if datatype(val, "Num")=.false then     -- a string in hand, get its corresponding numeric value
      val=self~at:super(val)

   -- retrieve Java object representing enum-value and return it
   return box("Long", val)                -- box it to a Long, will allow all values


::method "[]"        -- intercept []-message
--   say "   " self~string"/"self~"=="~c2x": []()-message,   arg(1)="pp(val)
   forward message ("AT")        -- re-route to "AT"-method; just to be sure that "[]" is treated
                                 -- in any case like "AT"

::method unknown
--   say "   " self~string"/"self~"=="~c2x": UNKNOWN()-message,   arg(1)="pp(arg(1)) "arg(2)="pp(arg(2))
   forward class (super) continue
--   say "   " "     result="pp(result)
   return self~at(result)        -- now let local AT-method do its thingy ...
*/




/* *************************************************************
   Class:      UNO_PROPERTIES
   Purpose:    make all UNO_PROPERTY definitions via an ooRexx directory object
   Remark:     - sending the property name will return its encoded UNOIDL string, which
                 needs to be parsed for further processing
   ************************************************************* */
::CLASS UNO_PROPERTIES SUBCLASS UNO_DIRLIKE -- PUBLIC
::method setup       -- check and setup directory object

    if self~unoidl_typeName<>"" then                  -- of correct UNO type ?
       return .false

    encodedProps=self~unoidl_name                     -- get argument
    parse var encodedProps . "|" type "|" .           -- check for encoded properties
    if type<>"UNO_PROPERTY" then
       return .false

    self~unoidl_name=""                               -- indicate no UNOIDL name available
    self~unoidl_typeName="UNO_PROPERTY"               -- indicate kind of directory
    self~unoidl_className=""                          -- indicate kind of directory

    self~unoidl_definition=encodedProps               -- save encoded property string

    tmpArr=encodedProps~makearray(" ")

    do i=1 to tmpArr~items                            -- iterate over CONSTANT values
       parse value tmpArr[i] with name "|" .
       self~setentry(name, tmpArr[i])                 -- save encoded property info
    end
    return .true                                      -- everything worked out o.k.



  -- wrap argument into an UNO-proxy object, if it is a BSF object, else return argument unchanged
::routine UNO.wrap public
  if arg(1)~class=.UNO_proxy then return arg(1) -- already wrapped up

  if arg(1)~hasmethod("BSF.EXIT") then
     return .UNO_proxy~new(arg(1)) -- o.k. a BSF object

  o=bsf.wrap(arg(1))    -- hmm, maybe not yet wrapped up as a BSF-object
  if o~hasmethod("BSF.EXIT") then
     return .UNO_proxy~new(o) -- o.k. a BSF object

  return arg(1)                                 -- return argument as is (unwrapped)


   -- convert a file-name with path to URL notation as needed by UNO
::routine ConvertToUrl public    -- named after {Star|OOo}Basic
  parse arg str
  return (iif(left(str,1)="/", "file://", "file:///") || encodeURL(str))~changestr("\", "/")


   -- convert a file-name with path encoded in the URL notation to an operating system compliant file with path
::routine ConvertFromUrl public  -- named after {Star|OOo}Basic
  parse arg str

  new=decodeURL(str)             -- decode URL (replace escapes with the escaped chars themselves


  if left(new,5)="file:" then
  do
     dosLike= (substr(new, 10,1)=":")  -- if 10th char a colon, assume DOS-like filename

     if dosLike=.true then
        new=substr(new, 9)
     else      -- leave initial slash
        new=substr(new, 8)

     if .UNO~hasentry("FILE.SEPARATOR")=.false then   -- not yet in .UNO directory, hence add it
        .UNO~file.separator=.java.lang.system~getProperty("file.separator")   -- get info from Java

      -- if a DOS-like path and on a DOS-like system, translate slashes to backslashes
     if dosLike=.true | .UNO~file.separator="\" then
        return changestr("/", new, "\")
  end
  return new



::routine encodeURL public
  parse arg str

  new=changestr("%", str, "%25")    -- escape escape char %

   -- determine chars in URLs that do not get encoded
  url.chars="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789/\?:=%'"".,+-"

  pos=verify(new, url.chars)     -- chars to escape in string?
  do while pos<>0
     char=substr(new, pos, 1)    -- extract char not in reference
     new=new~changestr(char, "%"||char~c2x)  -- escape character
     pos=verify(new, url.chars)
  end
  return new




   -- re-create chars from escape-sequences led in by % followed by two-digit hex value
::routine decodeUrl public    -- %20 --> " "
  parse arg str

  new=""
  do while str<>""
     parse var str before "%" +1 hexValue +2 str -- extract hex value
     new=new || before || hexValue~x2c
  end
  return new




   -- interface to UnoRuntime.areSame(o1, o2)
::routine uno.areSame public
   use arg o1, o2
   return .uno~UnoRuntime~areSame(o1, o2)     /* true if both reference the same UNO object   */


/* Pass-through routines for "org.oorexx.uno.RgfReflectUNO", wrapping up
   returned UNO objects into UNO_PROXY ooRexx objects */
   -- interface to RgfReflectUNO.findInterfaceWithMember(): can be a blank delimited string or an interface object
::routine uno.findInterfaceWithMember public
   use arg o, name, bREturnString, howMany

   if datatype(bReturnString, "Boolean")<>.true then
      bReturnString=.true                       -- default to return a blank delimited string

   if datatype(howMany, "Whole")<>.true then
      howMany=-1                                -- default to return all matches

      -- wrap up as an UNO proxy, if an object got returned
   return uno.wrap( .uno~rgfReflectUNO~findInterfaceWithMember( o, name, bReturnString, howMany, .uno~bExtendSearch) )


   -- interface to RgfReflectUNO.getDefinition(): returns always a blank delimited string
::routine uno.getDefinition public
   use arg o
   return .uno~rgfReflectUNO~getDefinition(o)


   -- interface to RgfReflectUNO.getInterfacesViaReflection(): returns always a blank delimited string
::routine uno.getInterfaceNamesViaReflection public
   use arg o
   return .uno~rgfReflectUNO~getInterfaceNamesViaReflection(o)


   -- interface to RgfReflectUNO.getProperties(): returns always a blank delimited string
::routine uno.getProperties public
   use arg o
   return .uno~rgfReflectUNO~getProperties(o)


   -- interface to RgfReflectUNO.getServicesViaReflection(): returns always a blank delimited string
::routine uno.getServiceNamesViaReflection public
   use arg o
   return .uno~rgfReflectUNO~getServiceNamesViaReflection(o)


   -- interface to RgfReflectUNO.getType(): always a string
::routine uno.getTypeName public
   use arg o
   return .uno~rgfReflectUNO~getTypeName(o)


   -- interface to RgfReflectUNO.getXTypeProviderTypes(): returns always a blank delimited string
::routine uno.getXTypeProviderTypeNames public
   use arg o
   return .uno~rgfReflectUNO~getXTypeProviderTypeNames(o)


   -- interface to RgfReflectUNO.queryInterfaceName(): always a string
::routine uno.queryInterfaceName public
   use arg o, name
   return .uno~rgfReflectUNO~queryInterfaceName(o, name, .uno~bExtendSearch)-- always a string


   -- interface to RgfReflectUNO.queryInterfaceObjectByName(): returns the interface object or .nil
::routine uno.queryInterfaceObjectByName public
   use arg o, name
      -- wrap up as an UNO proxy, if an object got returned
   return uno.wrap( .uno~rgfReflectUNO~queryInterfaceObjectByName(o, name, .uno~bExtendSearch) )


   -- interface to RgfReflectUNO.queryServiceName(): always a string
::routine uno.queryServiceName public
   use arg o, name
   return .uno~rgfReflectUNO~queryServiceName(o, name)


   -- "pretty print" strings from RgfReflectUNO: uses a blank to create an array from the string
   --                first entry is untoched, all others are delimited with CR-LF-TAB for indentation
::routine ppd public
  parse arg debugString, delimiter

  if delimiter=="" then delimiter=" "  -- default to a blank as delimiter

  arr=debugString~makearray(delimiter)
  tab="09"x
  crlf="0d0a"x
  crlftab=crlf||tab
  tmpStr=arr[1]
  do i=2 to arr~items
     tmpStr=tmpStr || crlftab || arr[i]
  end

  return tmpStr



::routine uno.getScriptContext public
return uno.wrap(bsf.lookupBean("OOo.xScriptContext"))



::routine uno.getScriptContextVersion public
return bsf.lookupBean("OOo.ooRexxScriptProvider.version")



